library(mvtnorm)
library(nlme)
library(verification)
library(gee)
library(numDeriv)

SRE=function(Y.long,TIMEF,TIMER,DD,ID,TR,Kp,subs){
## Y.long long version of outcome
## TIMEF fixed effects design matrix, with intercept
## TIMER random effects design matrix, with intercept
## DD gold standard (length N)
## ID clustering variable
## TR vector indicating the training set
## Kp time point for prediction, i.e., c(1,4.5,4.5^2)
## subs subset of time points, i.e., c(2,3,4), the last three time for prediction
## only works for balanced design - equal number of observations at
## fixed time points
NN=length(table(ID))
KK=max(table(ID))
MM=length(table(ID[TR==1]))
Y=Y.long[TR==1]
timeF=TIMEF[TR==1,]
timeR=TIMER[TR==1,]
id=ID[TR==1]
fitA=try(lme(Y~-1+timeF,random=~-1+timeR|id,control=lmeControl(maxIter=10000,opt="optim"),method="ML"))
if(length(fitA)<=5) return(NULL)
tempX=TIMEF[ID==1,][subs,]
tempZ=TIMER[ID==1,][subs,]

Sigmab=getVarCov(fitA,individual=1)
Ri=getVarCov(fitA,individual=1,type="conditional")[[1]][subs,subs]
Vi=Ri+tempZ%*%Sigmab%*%t(tempZ)
Wi=solve(Vi)

Vareb=Sigmab-Sigmab%*%t(tempZ)%*%Wi%*%tempZ%*%Sigmab

vv=c(Kp%*%Vareb%*%Kp)
Dum1=Sigmab%*%t(tempZ)%*%solve(Vi)
mu.fix=c(TIMEF%*%fitA$coef$fixed)
EB=c(Kp%*%Dum1%*%matrix(Y.long-mu.fix,nrow=KK)[subs,])
tr.index=(unique(ID[TR==1]))
Ans=glm(DD[tr.index]~EB[tr.index],family=binomial(link="probit"))
A2=Ans$coef[2]
A1=Ans$coef[1]
if (vv*A2^2<1){

A2=Ans$coef[2]/sqrt(1-vv*Ans$coef[2]^2)
A1=Ans$coef[1]*sqrt(1+A2^2*vv)
}
Pred=pnorm((A1+A2*EB)/sqrt(1+A2^2*vv))
list(marker=EB,risk=Pred,coef=c(A1,A2))
}

LR=function(Y.long,TIMEF,TIMER,DD,ID,TR){
NN=length(table(ID))
KK=max(table(ID))
MM=length(table(ID[TR==1]))
Y=Y.long[TR==1]
timeF=TIMEF[TR==1,]
timeR=TIMER[TR==1,]
id=ID[TR==1]
dd=DD[unique(id)]
longd=rep(dd,times=table(id))
Y1=Y[longd==1]
Y0=Y[longd==0]
timeF0=timeF[longd==0,]
timeF1=timeF[longd==1,]
timeR0=timeR[longd==0,]
timeR1=timeR[longd==1,]
id0=id[longd==0]
id1=id[longd==1]
fit0=try(lme(Y0~-1+timeF0,random=~-1+timeR0|id0,control=lmeControl(maxIter=10000,opt="optim"),method="ML"))
fit1=try(lme(Y1~-1+timeF1,random=~-1+timeR1|id1,control=lmeControl(maxIter=10000,opt="optim"),method="ML"))
if(length(fit0)<=5) return(NULL)
if(length(fit1)<=5) return(NULL)

DELTA0=getVarCov(fit0,individual=1)
DELTA1=getVarCov(fit1,individual=1)

thv0=c(diag(DELTA0),DELTA0[1,2],DELTA0[1,3],DELTA0[2,3])
the0=fit0$sigma^2
th0=fit0$coef$fixed
th.est0=c(th0,thv0,the0)

thv1=c(diag(DELTA1),DELTA1[1,2],DELTA1[1,3],DELTA1[2,3])
the1=fit1$sigma^2
th1=fit1$coef$fixed
th.est1=c(th1,thv1,the1)

fcoef1=fit1$coef$fixed
fcoef0=fit0$coef$fixed
risk=lrcom=TVAR=rep(0,NN)

HES0=hessian(likx,th.est0,wideY=t(matrix(Y0,nrow=KK)),TIMEF=timeF0,TIMER=timeR0,method.args=list(d=0.001))
VAR0=(-solve(HES0))
HES1=hessian(likx,th.est1,wideY=t(matrix(Y1,nrow=KK)),TIMEF=timeF1,TIMER=timeR1,method.args=list(d=0.001))
VAR0=(-solve(HES0))
VAR1=(-solve(HES1))
pD.est=mean(dd)
VARD=(pD.est*(1-pD.est))/length(DD)

for(i in 1:NN){
tempX=TIMEF[ID==i,]
tempZ=TIMER[ID==i,]
csize=nrow(tempX)
Ri1=diag(csize)*fit1$sigma^2
Vi1=Ri1+tempZ%*%DELTA1%*%t(tempZ)
Ri0=diag(csize)*fit0$sigma^2
Vi0=Ri0+tempZ%*%DELTA0%*%t(tempZ)
mu1=c(tempX%*%fcoef1)
mu0=c(tempX%*%fcoef0)
TT=Y.long[ID==i]
lrcom[i]=t(TT-mu0)%*%solve(Vi0)%*%c(TT-mu0)- t(TT-mu1)%*%solve(Vi1)%*%c(TT-mu1)
risk[i]=expit(lrcom[i]*0.5+0.5*(log(det(Vi0))-log(det(Vi1)))+
log(mean(DD)/(1-mean(DD))))
DV0=grad(likp,th.est0,TT=TT,tempX=tempX,tempZ=tempZ)
DV1=grad(likp,th.est1,TT=TT,tempX=tempX,tempZ=tempZ)
DVD=1/pD.est/(1-pD.est)
TVAR[i]=c(DV0%*%VAR0%*%(DV0))+ c(DV1%*%VAR1%*%(DV1))+DVD^2*VARD
}
list(risk=risk[(MM+1):NN],var.lo=TVAR[(MM+1):NN])
}


SREx=function(Y.long,TIMEF,TIMER,DD,ID,TR,Kp,subs,Xmat){
## Y.long long version of outcome
## TIMEF fixed effects design matrix, with intercept
## TIMER random effects design matrix, with intercept
## DD gold standard (length N)
## ID clustering variable
## TR vector indicating the training set
## Kp time point for prediction, i.e., c(1,4.5,4.5^2)
## subs subset of time points, i.e., c(2,3,4), the last three time for prediction
## only works for balanced design - equal number of observations at
## fixed time points
NN=length(table(ID))
KK=max(table(ID))
MM=length(table(ID[TR==1]))
Y=Y.long[TR==1]
timeF=TIMEF[TR==1,]
timeR=TIMER[TR==1,]
id=ID[TR==1]
fitA=try(lme(Y~-1+timeF,random=~-1+timeR|id,control=lmeControl(maxIter=10000,opt="optim"),method="ML"))
if(length(fitA)<=5) return(NULL)
tempX=TIMEF[ID==1,][subs,]
tempZ=TIMER[ID==1,][subs,]

Sigmab=getVarCov(fitA,individual=1)
Ri=getVarCov(fitA,individual=1,type="conditional")[[1]][subs,subs]
Vi=Ri+tempZ%*%Sigmab%*%t(tempZ)
Wi=solve(Vi)

Vareb=Sigmab-Sigmab%*%t(tempZ)%*%Wi%*%tempZ%*%Sigmab

vv=c(Kp%*%Vareb%*%Kp)
Dum1=Sigmab%*%t(tempZ)%*%solve(Vi)
mu.fix=c(TIMEF%*%fitA$coef$fixed)
EB=c(Kp%*%Dum1%*%matrix(Y.long-mu.fix,nrow=KK)[subs,])
tr.index=(unique(ID[TR==1]))
Ans=glm(DD[tr.index]~-1+EB[tr.index]+Xmat[tr.index,],family=binomial(link="probit"))
A2=Ans$coef[1]
A1=Ans$coef[-1]
if (vv*A2^2<1){
A2=Ans$coef[1]/sqrt(1-vv*Ans$coef[1]^2)
A1=Ans$coef[-1]*sqrt(1+A2^2*vv)
}
Pred=c(pnorm((Xmat%*%A1+A2*EB)/sqrt(1+A2^2*vv)))
list(marker=EB,risk=Pred,coef=c(A1,A2))
}


LRx=function(Y.long,TIMEF,TIMER,DD,ID,TR,Xmat){
NN=length(table(ID))
KK=max(table(ID))
MM=length(table(ID[TR==1]))
Y=Y.long[TR==1]
timeF=TIMEF[TR==1,]
timeR=TIMER[TR==1,]
id=ID[TR==1]
dd=DD[unique(id)]
longd=rep(dd,times=table(id))
Y1=Y[longd==1]
Y0=Y[longd==0]
timeF0=timeF[longd==0,]
timeF1=timeF[longd==1,]
timeR0=timeR[longd==0,]
timeR1=timeR[longd==1,]
id0=id[longd==0]
id1=id[longd==1]
fit0=try(lme(Y0~-1+timeF0,random=~-1+timeR0|id0,control=lmeControl(maxIter=10000,opt="optim"),method="ML"))
fit1=try(lme(Y1~-1+timeF1,random=~-1+timeR1|id1,control=lmeControl(maxIter=10000,opt="optim"),method="ML"))
if(length(fit0)<=5) return(NULL)
if(length(fit1)<=5) return(NULL)

DELTA0=getVarCov(fit0,individual=1)
DELTA1=getVarCov(fit1,individual=1)

thv0=c(diag(DELTA0),DELTA0[1,2],DELTA0[1,3],DELTA0[2,3])
the0=fit0$sigma^2
th0=fit0$coef$fixed
th.est0=c(th0,thv0,the0)

thv1=c(diag(DELTA1),DELTA1[1,2],DELTA1[1,3],DELTA1[2,3])
the1=fit1$sigma^2
th1=fit1$coef$fixed
th.est1=c(th1,thv1,the1)

fcoef1=fit1$coef$fixed
fcoef0=fit0$coef$fixed
risk=lrcom=TVAR=rep(0,NN)

HES0=hessian(likx,th.est0,wideY=t(matrix(Y0,nrow=KK)),TIMEF=timeF0,TIMER=timeR0,
method.args=list(d=0.001))
VAR0=(-solve(HES0))
HES1=hessian(likx,th.est1,wideY=t(matrix(Y1,nrow=KK)),TIMEF=timeF1,TIMER=timeR1,
method.args=list(d=0.001))

VAR1=(-solve(HES1))

fitD=glm(DD[unique(id)]~-1+Xmat[unique(id),],family="binomial")
pD.lo=Xmat%*%fitD$coef
VARD=summary(fitD)$cov.unscaled

for(i in 1:NN){
tempX=TIMEF[ID==i,]
tempZ=TIMER[ID==i,]
csize=nrow(tempX)
Ri1=diag(csize)*fit1$sigma^2
Vi1=Ri1+tempZ%*%DELTA1%*%t(tempZ)
Ri0=diag(csize)*fit0$sigma^2
Vi0=Ri0+tempZ%*%DELTA0%*%t(tempZ)
mu1=c(tempX%*%fcoef1)
mu0=c(tempX%*%fcoef0)
TT=Y.long[ID==i]
lrcom[i]=t(TT-mu0)%*%solve(Vi0)%*%c(TT-mu0)- t(TT-mu1)%*%solve(Vi1)%*%c(TT-mu1)
risk[i]=expit(lrcom[i]*0.5+0.5*(log(det(Vi0))-log(det(Vi1)))+pD.lo[i])
DV0=grad(likpx,th.est0,TT=TT,tempX=tempX,tempZ=tempZ)
DV1=grad(likpx,th.est1,TT=TT,tempX=tempX,tempZ=tempZ)
TVAR[i]=c(DV0%*%VAR0%*%(DV0))+ c(DV1%*%VAR1%*%(DV1))+c(Xmat[i,]%*%VARD%*%Xmat[i,])
}
list(risk=risk[(MM+1):NN],var.lo=TVAR[(MM+1):NN])
}

#utility-functions

expit=function(x) 1/(1+exp(-x))

likx=function(th.est,wideY,TIMEF,TIMER){
pl=ncol(TIMEF)
ql=ncol(TIMER)
th=th.est[1:pl]
thv=th.est[(pl+1):(pl+ql*(ql+1)/2)]
the=th.est[pl+ql*(ql+1)/2+1]
mu=c(TIMEF%*%th)
mu.w=t(matrix(mu,ncol(wideY),nrow(wideY)))
DELTA=diag(thv[1:3])
DELTA[1,2]=DELTA[2,1]=thv[4]
DELTA[1,3]=DELTA[3,1]=thv[5]
DELTA[2,3]=DELTA[3,2]=thv[6]
csize=ncol(wideY)
Ri=diag(csize)*the
log.lik=rep(0,nrow(wideY))
for(i in 1:nrow(wideY)){
mui=mu.w[i,]
Vi=Ri+TIMER[(i-1)*csize+(1:csize),]%*%DELTA%*%t(TIMER[(i-1)*csize+(1:csize),])
log.lik[i]=(dmvnorm(wideY[i,],mui,Vi,log=T))
}
sum(log.lik)
}

likp=function(th.est,TT,tempX,tempZ){
th=th.est[1:3]
thv=th.est[4:9]
the=th.est[10]
mu=c(tempX%*%th)
DELTA=diag(thv[1:3])
DELTA[1,2]=DELTA[2,1]=thv[4]
DELTA[1,3]=DELTA[3,1]=thv[5]
DELTA[2,3]=DELTA[3,2]=thv[6]
csize=nrow(tempX)
Ri=diag(csize)*the
Vi=Ri+tempZ%*%DELTA%*%t(tempZ)
dmvnorm(TT,mu,Vi,log=T)
}

likpx=function(th.est,TT,tempX,tempZ){
pl=ncol(tempX)
ql=ncol(tempZ)
th=th.est[1:pl]
thv=th.est[(pl+1):(pl+ql*(ql+1)/2)]
the=th.est[pl+ql*(ql+1)/2+1]
mu=c(tempX%*%th)
DELTA=diag(thv[1:3])
DELTA[1,2]=DELTA[2,1]=thv[4]
DELTA[1,3]=DELTA[3,1]=thv[5]
DELTA[2,3]=DELTA[3,2]=thv[6]
csize=nrow(tempX)
Ri=diag(csize)*the
Vi=Ri+tempZ%*%DELTA%*%t(tempZ)
dmvnorm(TT,mu,Vi,log=T)
}

